*********************
********************* application to 7 variable VAR in Smets-Wouters variables, OLS estimation
********************* generating forecasts and forecast tests, conditional and unconditional 
********************* Todd Clark, Federal Reserve Bank of Cleveland, todd.clark@clev.frb.org
********************* 

dis %dateandtime()
dis %ratsversion()

*********************
********************* BASIC SETUP & PARAMETER ENTRY
*********************
comp condpos = 1  ;* this is the position in the VAR variable vector for the variable on which forecasts will be conditioned
                   * 1 conditions on the 1st variable, 2 conditions on the 2nd variable, etc.

comp nboot = 1999   ;* number of bootstrap replications (1999 used in the paper's results)

comp styr = 1947
cal(q) styr:1
comp eststpt = styr:1	;*earliest period with data in the file
comp endsmpl = 2007:4	;*last period of forecast evaluation sample.  Note: forecasts are dated by observation date, not forecast origin.

comp stsmpl = 1961:1  ;* earliest starting point for VAR models
comp fcst = 1991:1    ;* first forecast origin
comp blocksize = 40   ;* block size using in moving block bootstrap of residuals
comp scheme = 0       ;* 0 for recursive forecasts, 1 for rolling.  In the paper, we use the recursive scheme.
comp rollwsize = 80   ;* size of rolling window if scheme is rolling

comp nvar = 7	;* number of variables in the VAR
comp lags = 4	;* fixed lag order to use in VAR

comp [vec[int]] horz = ||1,2,4,8||   ;* forecast horizons, of 1, 2, 4, and 8 quarters
comp nhorz = %rows(horz)
comp maxh = horz(nhorz)    ;* longest possible forecast horizon (furthest period out)
comp numcond = 8           ;* number of periods over which conditions are imposed

comp endsmpl2 = endsmpl+maxh-1      ;* we need to pad the sample to deal with our forward looking calcs
all endsmpl+maxh   
smpl eststpt endsmpl+maxh

comp seedval = 4*128*%month(endsmpl)*%year(endsmpl)
seed seedval
dis seedval

*********************
********************* READING IN DATA
*********************
open data CMdata.xlsx
data(format=xlsx,org=col) / pce hours gdp invest hrlycomp pigdp popul_frbus itb3m ffr
close
** note: invest is a chain-weight index of real investment, not a dollar series (dollar series from BEA not available until 1990s)

** some data adjustments
set ffr eststpt 1959:4 = itb3m           ;* use 3-month T-bill in early years of sample (FFR was not available or market was thin)
set(scratch) hrlycomp = hrlycomp/pigdp   ;* real hourly comp = nominal/GDP price index

** form growth rates and per-capita growth rates
set(scratch) popgrowth = 400.*log(popul_frbus/popul_frbus{1})
dofor i = gdp pce invest hrlycomp hours pigdp
 set(scratch) i = 400.*log(i{0}/i{1})
 set %s('pc_'+%l(i)) = i{0} - popgrowth
end do i

*********************
********************* DATA TRANSFORMATIONS and setting up data vectors
*********************
dec vec[ser] y(nvar) yorig(nvar) actuals(nvar)

comp [vec[str]] varlabel = ||"GDP", "PCE", "INVEST", "WAGES", "HOURS", "INFL", "FFR"||

*** store data in vectors
comp cnt = 0
dofor i = pc_gdp pc_pce pc_invest hrlycomp pc_hours pigdp ffr
 comp cnt = cnt + 1
 set(scratch) y(cnt) = i{0}
 set yorig(cnt) stsmpl-lags endsmpl2 = y(cnt){0} ;* create copy of series, as we will need them in bootstrap resampling
end do i

********************************
******************************** stuff for estimation of VAR
********************************
** model defined for use below (need to declare before procedures, which use varmodel)
system(model=varmodel)
variables y
lags 1 to lags
det constant
end(system)

** procedures
sou(noecho) VARestloop.src            ;* procedure for estimating VAR over forecast sample and storing time series of param. estimates
sou(noecho) VARcondfc.src             ;* procedure for forming unconditional and conditional forecasts from VAR
sou(noecho) pctiles2s.src             ;* function for computing bootstrap percentiles (2-sided)
sou(noecho) regressfcn.src            ;* function for robust regression approach used in test statistics

********************************
******************************** stuff for forecast computation result storage
********************************  note that we compute forecasts for all horizons up to maxh but only do testing for the selected horizons called for in the vector horz
********************************
dec rec[ser] pointfc_u(nvar,maxh) pointfc_dls(nvar,maxh) err_u(nvar,maxh) err_dls(nvar,maxh) actualsmatrix(nvar,maxh)
clear(length=endsmpl+maxh) pointfc_u pointfc_dls err_u err_dls  

** now create series labels for Excel storage and define horizon-specific actuals used in forecast evaluation (used for various computations)
do h = 1,maxh
 do i = 1,nvar
  labels pointfc_u(i,h)
  # 'ufc_v'+%string(i)+'_h'+%string(h)
  labels pointfc_dls(i,h)
  # 'dlsfc_v'+%string(i)+'_h'+%string(h)
  labels actualsmatrix(i,h)
  # 'actual_v'+%string(i)+'_h'+%string(h)
  *
  set actualsmatrix(i,h) = y(i){0}
 end do i
end do h

****************************************
**************************************** stuff for forecast testing and storage
****************************************
comp [vec[str]] testlabels = ||'bias', 'efficiency', 'DM equal MSE'||
comp tottests = %rows(testlabels)
comp [vec[str]] typelabels = ||'uncond.', 'condit.'||

comp [rec] normtests = %zeros(tottests,nvar*nhorz)  ;* for storage of sample stats
dec rec[vec] regcoefs(tottests,nvar*nhorz)          ;* for storage of test regression coefficients
dec ser[rec] artbsstats					         	;* for storage of bootstrap test stats
gset artbsstats 1 nboot = %zeros(tottests,nvar*nhorz)

dec rec[rec] bscritvals(tottests,nvar*nhorz)   ;* for storage of bootstrap critical values
dec vec[series] bsresids(nvar) fsresids(nvar)  ;* residuals we use in bootstrap

** bandwidth for HAC estimators
dec vec[int] hacbw(nhorz)
do h = 1,nhorz
 comp hacbw(h) = horz(h)-1
end do h

** for tracking average (across MC draws) MSEs
comp [rec] mses = %zeros(2,nvar*nhorz)

****************************************
**************************************** generate forecasts and sample test statistics
****************************************
** if we use rolling scheme, reset the starting point of estimation to reflect chosen window size
if scheme==1
 comp stsmpl = (fcst-1)-rollwsize+1
endif
dis %datelabel(stsmpl)  %datelabel(endsmpl2)

****************************** estimate VAR over full sample, for use below
****************************** full sample is defined to go through date of last condition used in sample
estimate(print,noftests,residuals=fsresids) stsmpl endsmpl2  ;* we will use coefs and residuals in bootstrap sampling
comp [rec] fscoefres = %modelgetcoeffs(varmodel)
comp [symm] fssigmares = %sigma

****************************** compute population-level difference in MSEs
impulse(model=varmodel,results=impfunc,noprint,decomp=%decomp(fssigmares)) * maxh
comp [rec] bigrmat = %zeros(nvar*maxh,nvar*maxh)
comp [rec] mamatrix = %zeros(nvar,nvar)
comp [rec] rtilde = %zeros(numcond,nvar*maxh)
do hh = 1,maxh
 ewise mamatrix(i,j) = impfunc(i,j)(hh)
 do kk = hh,maxh
  comp %psubmat(bigrmat,(kk-1)*nvar+1,(kk-hh)*nvar+1,mamatrix)
 end do kk
end do hh
do hh = 1,numcond
 comp %psubmat(rtilde,hh,1,%xrow(bigrmat,(hh-1)*nvar+condpos))
end do hh
comp [vec] popmsediffvec = %xdiag(%mqform(inv(rtilde*tr(rtilde)),(rtilde*tr(bigrmat)))) 
comp [rec] popmsediff = %vectorect(popmsediffvec,nvar)  ;* puts pop-level MSE diffs in matrix of dimension nvar x maxh
comp keepmsediff = popmsediff

****************************** estimation and forecasting:  OOS
*** estimation, to produce time series of model estimates over forecast sample
@VARloop(scheme=scheme) y stsmpl fcst endsmpl lags coefres sigmares

*** loop over time to form unconditional and conditional forecasts
do time = fcst,endsmpl
 *** define path of policy variable used in conditional forecasting
 clear(length=endsmpl+maxh) polpath
 set polpath time time+maxh-1 = y(condpos){0}
 *** generate forecasts (would need to change number of conditions here if maxh is changed to not = 8)
 @VARcondfc y lags coefres(time) sigmares(time) time maxh numcond fcast_u fcast_dls
 # condpos time-1+1 polpath(time-1+1)
 # condpos time-1+2 polpath(time-1+2)
 # condpos time-1+3 polpath(time-1+3)
 # condpos time-1+4 polpath(time-1+4)
 # condpos time-1+5 polpath(time-1+5)
 # condpos time-1+6 polpath(time-1+6)
 # condpos time-1+7 polpath(time-1+7)
 # condpos time-1+8 polpath(time-1+8)

 *** store forecasts, after averaging as appropriate
 do i = 1,nvar
  do h = 1,maxh
   comp pointfc_u(i,h)(time-1+h) = fcast_u(i)(time-1+h)
   comp pointfc_dls(i,h)(time-1+h) = fcast_dls(i)(time-1+h)
  end do h
 end do i
end do time

****************************** form some forecast time series not formed above but needed in efficiency regressions
do h = 1,maxh
 do i = 1,nvar
  set(scratch) err_u(i,h) fcst endsmpl = actualsmatrix(i,h){0} - pointfc_u(i,h){0}
  set(scratch) err_dls(i,h) fcst endsmpl = actualsmatrix(i,h){0} - pointfc_dls(i,h){0}
 end do i
end do h

****************************** compute and store MSEs for all variables and forecasts
do i = 1,nvar
 do hh = 1,nhorz
  comp jj = 0
  dofor j = err_u(i,horz(hh)) err_dls(i,horz(hh)) 
   comp jj = jj + 1
   sstats(mean) fcst+horz(hh)-1 endsmpl j{0}^2.>>mses(jj,(i-1)*nhorz+hh)
  end do j
 end do hh
end do i

****************************** form bias, efficiency, and MSE tests
****************************** note that we define the efficiency regression so the test coefficient of interest is the first (not last) coefficient
do i = 1,nvar
 if i==condpos       ;* we can't form tests for variable on which we are conditioning; we skip that position
  next
 do hh = 1,nhorz
  smpl fcst+horz(hh)-1 endsmpl
  comp usebw = hacbw(hh)

  *********** equal MSE test, uncond vs DLS
  set lossdiff = (err_u(i,horz(hh)){0}^2. - err_dls(i,horz(hh)){0}^2.) - popmsediff(i,horz(hh))
  @regressfcn(noprint) lossdiff usebw
  # constant
  comp normtests(tottests,(i-1)*nhorz+hh) = %tstats(1)
  comp regcoefs(tottests,(i-1)*nhorz+hh) = %beta

  set lossdiff = err_dls(i,horz(hh)){0}

  *********** bias test
  comp jj = 1
  @regressfcn(noprint) lossdiff usebw
  # constant
  comp normtests(jj,(i-1)*nhorz+hh) = %tstats(1)
  comp regcoefs(jj,(i-1)*nhorz+hh) = %beta

  *********** efficiency test
  comp jj = 2
  @regressfcn(noprint) lossdiff usebw
  # pointfc_dls(i,horz(hh)){0} constant
  comp normtests(jj,(i-1)*nhorz+hh) = %tstats(1) 
  comp regcoefs(jj,(i-1)*nhorz+hh) = %beta

 end do hh
end do i

dis %dateandtime()

******************************
****************************** BOOTSTRAP
******************************

do bb = 1,nboot
  ****************************** data generation, estimation and forecasting
  comp %modelsetcoeffs(varmodel,fscoefres)  ;* note that coef estimates used to produce bs data are the full sample coef estimates
  *** create artificial data, and re-assign actuals used in forecast evaluation
  comp initpoint = %raninteger(stsmpl,endsmpl2)
  boot(block=blocksize) entry stsmpl endsmpl2                  ;* moving block bootstrap of residuals (overlapping blocks)
  do i = 1,nvar
   set y(i) stsmpl-lags stsmpl-1 = yorig(i){stsmpl-initpoint}  ;* we depart from BJT in randomizing initial values instead of setting them to 0
   set bsresids(i) stsmpl endsmpl2 = fsresids(i)(entry(t))
   diff(center) bsresids(i) stsmpl endsmpl2                    ;* we depart from BJT in simply recentering the bootstrapped residuals (in lieu of their more proper recentering which is asymptotically irrelevant)
  end do i
  forecast(model=varmodel,paths,from=stsmpl,to=endsmpl2,results=y) *
  # bsresids

  clear(length=endsmpl+maxh) actualsmatrix
  do h = 1,maxh
   do i = 1,nvar
     set actualsmatrix(i,h) stsmpl endsmpl = y(i){0}
   end do i
  end do h

  ****************************** estimate BVAR over full sample, for use below
  estimate(noprint) stsmpl endsmpl2  ;* we will use coefs and residuals in bootstrap sampling
  comp [rec] bscoefres = %modelgetcoeffs(varmodel)
  comp [symm] bssigmares = %sigma

  ****************************** compute population-level difference in MSEs
  impulse(model=varmodel,results=impfunc,noprint,decomp=%decomp(bssigmares)) * maxh
  comp [rec] bigrmat = %zeros(nvar*maxh,nvar*maxh)
  comp [rec] mamatrix = %zeros(nvar,nvar)
  comp [rec] rtilde = %zeros(numcond,nvar*maxh)
  do hh = 1,maxh
   ewise mamatrix(i,j) = impfunc(i,j)(hh)
   do kk = hh,maxh
    comp %psubmat(bigrmat,(kk-1)*nvar+1,(kk-hh)*nvar+1,mamatrix)
   end do kk
  end do hh
  do hh = 1,numcond
   comp %psubmat(rtilde,hh,1,%xrow(bigrmat,(hh-1)*nvar+condpos))
  end do hh
  comp [vec] popmsediffvec = %xdiag(%mqform(inv(rtilde*tr(rtilde)),(rtilde*tr(bigrmat)))) 
  comp [rec] popmsediff = %vectorect(popmsediffvec,nvar)  ;* puts pop-level MSE diffs in matrix of dimension nvar x maxh

  ****************************** estimation and forecasting:  OOS
  *** estimation, to produce time series of model estimates over forecast sample
  @VARloop(scheme=scheme) y stsmpl fcst endsmpl lags coefres sigmares

  *** loop over time to form unconditional and conditional forecasts
  do time = fcst,endsmpl
   *** define path of policy variable used in conditional forecasting under policy shock approach
   clear(length=endsmpl+maxh) polpath
   set polpath time time+maxh-1 = y(condpos){0}

   *** generate forecasts (would need to change number of conditions here if maxh is changed to not = 8)
   @VARcondfc y lags coefres(time) sigmares(time) time maxh numcond fcast_u fcast_dls
   # condpos time-1+1 polpath(time-1+1)
   # condpos time-1+2 polpath(time-1+2)
   # condpos time-1+3 polpath(time-1+3)
   # condpos time-1+4 polpath(time-1+4)
   # condpos time-1+5 polpath(time-1+5)
   # condpos time-1+6 polpath(time-1+6)
   # condpos time-1+7 polpath(time-1+7)
   # condpos time-1+8 polpath(time-1+8)

   *** store forecasts, after averaging as appropriate
   do i = 1,nvar
    do h = 1,maxh
     comp pointfc_u(i,h)(time-1+h) = fcast_u(i)(time-1+h)
     comp pointfc_dls(i,h)(time-1+h) = fcast_dls(i)(time-1+h)
    end do h
   end do i
  end do time

  ****************************** form some forecast time series not formed above but needed in efficiency regressions
  do h = 1,maxh
   do i = 1,nvar
    set(scratch) err_u(i,h) fcst endsmpl = actualsmatrix(i,h){0} - pointfc_u(i,h){0}
    set(scratch) err_dls(i,h) fcst endsmpl = actualsmatrix(i,h){0} - pointfc_dls(i,h){0}
   end do i
  end do h

  ****************************** form bias, efficiency, and MSE tests
  ****************************** note that we define the efficiency regression so the test coefficient of interest is the first (not last) coefficient
  do i = 1,nvar
   if i==condpos       ;* we can't form tests for variable on which we are conditioning; we skip that position
    next
   do hh = 1,nhorz
    smpl fcst+horz(hh)-1 endsmpl
    comp usebw = hacbw(hh)

    *********** equal MSE test, uncond vs DLS
    set lossdiff = (err_u(i,horz(hh)){0}^2. - err_dls(i,horz(hh)){0}^2.) - popmsediff(i,horz(hh))
    @regressfcn lossdiff usebw
    # constant
    comp artbsstats(bb)(tottests,(i-1)*nhorz+hh) = %tstats(1)

    set lossdiff = err_dls(i,horz(hh)){0}

    *********** bias test
    comp jj = 1
    @regressfcn(noprint) lossdiff usebw
    # constant
    comp artbsstats(bb)(jj,(i-1)*nhorz+hh) = %tstats(1)

    *********** efficiency test
    comp jj = 2
    set lossdiff = err_dls(i,horz(hh)){0}
    @regressfcn(noprint) lossdiff usebw
    # pointfc_dls(i,horz(hh)){0} constant
    comp artbsstats(bb)(jj,(i-1)*nhorz+hh) = %tstats(1) 

   end do hh
  end do i

end do bb

dis %dateandtime()

**************************************** tabulation and storage of bootstrap critical values
smpl 1 nboot

do tt = 1,tottests
 do i = 1,nvar
  if i==condpos
   next
  do hh = 1,nhorz
    set statser 1 nboot = artbsstats(t)(tt,(i-1)*nhorz+hh) 
    comp bscritvals(tt,(i-1)*nhorz+hh) = pctile2s(statser,nboot)  ;* this will be a 2 by 2 matrix, with 10% (row 1) and 5% (row 2) crit values
  end do hh
 end do i
end do tt

****************************************
**************************************** display of MSEs and estimates of population difference in MSEs
****************************************

************** MSEs for all variables and forecasts
do i = 1,nvar
  dis ''
  do jj = 1,2  ;* uncond, condit
   dis ''
   do hh = 1,nhorz
    dis  'variable ' varlabel(i) @18 'h = ' horz(hh) @28 typelabels(jj) @46 'MSE = ' ####.### mses(jj,(i-1)*nhorz+hh) @66 'RMSE = ' ####.### mses(jj,(i-1)*nhorz+hh)^0.5
   end do hh
  end do jj
end do i

************** population differences in MSEs
do i = 1,nvar
  comp [vec] tempvec = %xrow(keepmsediff,i)
  dis varlabel(i) @18 ####.### tempvec
end do i

****************************************
**************************************** display of test results
****************************************
/* here are asymptotic critical values of relevance:
comp [vec] critvals = ||1.645, 1.96||  ;* st normal critical values for each block/table
comp [vec] critvals1s = ||1.282,1.645||  ;* st normal critical values for each block/table
comp [vec] chisqrcritvals = ||6.251,7.815||  ;* chi square critical values for each block/table, for Wald test (F-W joint) with 3 restrictions
*/

do i = 1,nvar
 if i==condpos
  next
 dis ''
 dis '*********************************** variable: ' varlabel(i)
 do hh = 1,nhorz
  dis ''
  dis '*********** h =' horz(hh) 
  do pos = 1,tottests
   dis ''
   dis 'test name = ' testlabels(pos)
   dis 'coef vec = ' @15 ####.### regcoefs(pos,(i-1)*nhorz+hh)
   dis 'test stat' @15 ####.### normtests(pos,(i-1)*nhorz+hh)
   comp [vec] tempvec = %xrow(bscritvals(pos,(i-1)*nhorz+hh),1)
   dis '10% cv' @10 ####.### tempvec
   comp [vec] tempvec = %xrow(bscritvals(pos,(i-1)*nhorz+hh),2)
   dis '5% cv' @10 ####.### tempvec
  end do pos
 end do hh
end do ii

****************************************
**************************************** building more convenient tables:  efficiency tests and MSE tests
****************************************
dec rec efficteststats(nvar,2*nhorz) MSEteststats(nvar,4*nhorz)

************* efficiency tests
do i = 1,nvar
 if i==condpos
  next
 do hh = 1,nhorz
  comp count = 0
  dofor pos = 2  ;* 3=M-Z efficiency test for condit forecasts
   comp count = count + 1

   comp teststat = normtests(pos,(i-1)*nhorz+hh)
   comp [vec] tempvec10 = %xrow(bscritvals(pos,(i-1)*nhorz+hh),1)
   comp [vec] tempvec5 = %xrow(bscritvals(pos,(i-1)*nhorz+hh),2)
   if teststat>tempvec5(2).or.teststat<tempvec5(1)
    comp pvalindic = 90.
   else if teststat>tempvec10(2).or.teststat<tempvec10(1)
    comp pvalindic = 80.
   else
    comp pvalindic = 20.

   comp efficteststats(i,(count-1)*2*nhorz+(hh-1)*2+1) = teststat
   comp efficteststats(i,(count-1)*2*nhorz+hh*2) = pvalindic

  end do pos
 end do hh
end do ii

************* MSE tests
comp pos = tottests  ;* MSE test, DLS vs uncond forecast
do i = 1,nvar
 if i==condpos
  next
 do hh = 1,nhorz

   comp teststat = normtests(pos,(i-1)*nhorz+hh)
   comp [vec] tempvec10 = %xrow(bscritvals(pos,(i-1)*nhorz+hh),1)
   comp [vec] tempvec5 = %xrow(bscritvals(pos,(i-1)*nhorz+hh),2)
   if teststat>tempvec5(2).or.teststat<tempvec5(1)
    comp pvalindic = 90.
   else if teststat>tempvec10(2).or.teststat<tempvec10(1)
    comp pvalindic = 80.
   else
    comp pvalindic = 20.
   comp MSEteststats(i,(hh-1)*4+1) = teststat
   comp MSEteststats(i,(hh-1)*4+2) = pvalindic
   comp MSEteststats(i,(hh-1)*4+3) = mses(1,(i-1)*nhorz+hh)-mses(2,(i-1)*nhorz+hh)
   comp MSEteststats(i,hh*4) = keepmsediff(i,horz(hh))

 end do hh
end do ii

************* now display these tables
************* Note:  the row pertaining to the variable conditioned on (row condpos) will display "NA" for all entries
/*
signif indicators:  90 for 5% signif, 80 for 10% signif, 20 for not signif at 10% or less

efficiency test table structure:
  M-Z test in cols 1,3,5,7 (horzs 1 through 8), signif indicators in cols 2,4,6,8

MSE test table structure:
  t-tests in cols 1,5,9,13 (horzs 1 through 8)
  signif indicators in cols 2,6,10,14
  actual (MSE-U - MSE-DLS) in cols 3,7,11,15
  popul diff in MSEs in cols 4,8,12,16
*/

dis ####.### efficteststats

dis ####.### MSEteststats

dis %dateandtime()
